module ST : sig
  type ('s, 'a) st
  val return : 'a -> ('s, 'a) st
  val bind : ('a -> ('s, 'b) st) -> ('s, 'a) st -> ('s, 'b) st

  type ('s, 'a) st_ref
  val st_ref : 'a -> ('s, ('s, 'a) st_ref) st
  val get : ('s, 'a) st_ref -> ('s, 'a) st
  val set : ('s, 'a) st_ref -> 'a -> ('s, unit) st

  type 'a pure = { run : 's . unit -> ('s, 'a) st }
  val run_st : 'a pure -> 'a
end = struct
  type ('s, 'a) st = 'a
  let return x = x
  let bind f m = f m

  type ('s, 'a) st_ref = 'a ref
  let st_ref, get, set = ref, (!), (:=)

  type 'a pure = { run : 's . unit -> ('s, 'a) st }
  let run_st st = st.run ()
end

let (>>=) m f = ST.bind f m


let test =
    ST.run_st { ST.run = fun () ->
      ST.st_ref 1 >>= fun r ->
      ST.get r >>= fun before ->
      ST.set r 2 >>= fun () ->
      ST.get r >>= fun after ->
      ST.return (before + after)
    };;



(* let test_escape_1 = *)
(*     ST.run_st { ST.run = fun () -> *)
(*       ST.st_ref 1 >>= fun r -> *)
(*       ST.return r *)
(*     } *)

let test_escape_2 =
    let hole = ref None in
    let f = fun () ->
      ST.st_ref 1 >>= fun r ->
      let () = hole := Some r in
      ST.return () in
    ST.run_st { ST.run = f };
    !hole
